/*
 *  pe_code_analyzer.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 8/30/05.
 *  Copyright 2005, 2006, 2007 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-01-16 permit the surfix of ".pdb"
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_code_analyzer.h"
#include "pe_print.h"

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;

void print_int_to_protein_code(struct pe_ud_code ** ic_tab)
{
	LONG p_num, c_ct;
	int n, t_ct;

	n = my_settings.res_code_length;
	p_num = (LONG) pow(2.0, (double) my_settings.res_code_length);

	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
"integer_to_U/D-sequence conversion table: Dump list starts (num_path = 2**%d = %ld)\n",
		n, p_num);
	if (ic_tab == NULL) {
		fprintf(STDOUT,
"intger_to_U/D-sequence conversion table: Memory not assigned.\n");
		fprintf(STDOUT, "\n");
		return;
	}
	for (c_ct = 0; c_ct < p_num; c_ct++) {
		/* U/D sequence */
		fprintf(STDOUT, "int_to_code[%8ld] = ", c_ct);
		for (t_ct = 0; t_ct < n - 1; t_ct++) {
			if (ic_tab[c_ct]->ud_seq[t_ct] == C_DOWN)
				fprintf(STDOUT, "D-");
			else
				fprintf(STDOUT, "U-");
		}
		if (ic_tab[c_ct]->ud_seq[t_ct] == C_DOWN)
			fprintf(STDOUT, "D");
		else
			fprintf(STDOUT, "U");
		fprintf(STDOUT,
			" (inv_code = %ld, up_num = %d)\n",
			ic_tab[c_ct]->inv_code, ic_tab[c_ct]->up_num);
	}
	fprintf(STDOUT,
"integer_to_U/D-sequence conversion table: Dump list ends.\n");
	fprintf(STDOUT, "\n");
	return;
}

void inv_code_check(struct pe_ud_code ** ic_tab)
{
	LONG p_num, c_ct;
	LONG inv_code, inv_ct;
	int n, t_ct;

	n = my_settings.res_code_length;
	p_num = (LONG) pow(2.0, (double) my_settings.res_code_length);
	inv_ct = 0;

	for (c_ct = 0; c_ct < p_num; c_ct++) {
		inv_code = ic_tab[c_ct]->inv_code;
		if (inv_code >= 0) {
			for (t_ct = 0; t_ct < n; t_ct++) {
				if (ic_tab[c_ct]->ud_seq[t_ct] == ic_tab[inv_code]->ud_seq[t_ct]) {
					fprintf(STDERR, "inv_code_check(): Error in %d-th tile (code = %ld <=> inv_code = %ld)\n", t_ct, c_ct, inv_code);
					exit(1);
				}
			}
			inv_ct++;
		}
	}
	if (inv_ct * 2 != p_num) {
		fprintf(STDERR, "inv_code_check(): Error #inverted_code = %ld <=> #all_code = %ld)\n", inv_ct, p_num);
		exit(1);
	}
	return;
}

struct pe_ud_code **setup_int_to_code_table(LONG p_num)
{
	struct pe_ud_code **ic_tab;
	LONG c_ct;
	LONG r_code, tmp;
	LONG inv_code, inv_step;
	int n, t_ct;
	int up_ct;

	n = my_settings.res_code_length;
	/* Setup the `r_code --> U/D-sequence conversion' table */
	ic_tab = (struct pe_ud_code **) malloc(p_num *
					       sizeof(struct pe_ud_code *));
	if (ic_tab == NULL) {
		fprintf(STDERR,
"setup_int_to_code_table(): Error ic_tab not allocated: p_num= %ld.\n",
			p_num);
		exit(1);
	}
	for (c_ct = 0; c_ct < p_num; c_ct++) {
		ic_tab[c_ct] = (struct pe_ud_code *)
				malloc(sizeof(struct pe_ud_code));
		ic_tab[c_ct]->ud_seq = (PE_CODE *) malloc(n * sizeof(PE_CODE));
		ic_tab[c_ct]->inv_code = -1;
		ic_tab[c_ct]->up_num = 0;
		r_code = c_ct;
		inv_code = 0;
		inv_step = 1;
		up_ct = 0;
		/* c_ct = code[0]*2**(n-1) +  code[1]*2**(n-2) +  ...
			+ code[n-2]*2 + code[n-1] */
		for (t_ct = n - 1; t_ct >= 0; t_ct--) {	
			tmp = r_code % 2;
			if (tmp == 0) {
				ic_tab[c_ct]->ud_seq[t_ct] = C_DOWN;
				inv_code = inv_code + inv_step;
			} else {
				ic_tab[c_ct]->ud_seq[t_ct] = C_UP;
				up_ct++;
			}
			r_code = r_code / 2;
			inv_step = inv_step * 2;
		}
		ic_tab[c_ct]->up_num = up_ct;
		
		if (up_ct * 2 > n) {
			ic_tab[c_ct]->inv_code = inv_code;
		} else if (up_ct * 2 == n) {
			if (inv_code < c_ct)
				ic_tab[c_ct]->inv_code = inv_code;
			else
				ic_tab[c_ct]->inv_code = -inv_code;
		} else {
			ic_tab[c_ct]->inv_code = -inv_code;
		}
		if (DEBUG) {
			if (r_code != 0) {
				fprintf(STDERR,
"setup_int_to_code_table(): Error int-to-code conversion (r_code= %ld)\n",
					r_code);
				exit(1);
			}
		}
	}
	if (DEBUG)
		inv_code_check(ic_tab);
	return ic_tab;
}

/* Currently not used */
void int_to_resname(int res_int, char *r_name)
{
	switch (res_int) {
		case (L_ALA):strncpy(r_name, "ALA\0", 4);
		return;
	case (L_ASP):
		strncpy(r_name, "ASP\0", 4);
		return;
	case (L_ARG):
		strncpy(r_name, "ARG\0", 4);
		return;
	case (L_ASN):
		strncpy(r_name, "ASN\0", 4);
		return;
	case (L_ASX):
		strncpy(r_name, "ASX\0", 4);
		return;
	case (L_CYS):
		strncpy(r_name, "CYS\0", 4);
		return;
	case (L_GLY):
		strncpy(r_name, "GLY\0", 4);
		return;
	case (L_GLU):
		strncpy(r_name, "GLU\0", 4);
		return;
	case (L_GLN):
		strncpy(r_name, "GLN\0", 4);
		return;
	case (L_GLX):
		strncpy(r_name, "GLX\0", 4);
		return;
	case (L_HIS):
		strncpy(r_name, "HIS\0", 4);
		return;
	case (L_ILE):
		strncpy(r_name, "ILE\0", 4);
		return;
	case (L_LEU):
		strncpy(r_name, "LEU\0", 4);
		return;
	case (L_LYS):
		strncpy(r_name, "LYS\0", 4);
		return;
	case (L_MET):
		strncpy(r_name, "MET\0", 4);
		return;
	case (L_PHE):
		strncpy(r_name, "PHE\0", 4);
		return;
	case (L_PRO):
		strncpy(r_name, "PRO\0", 4);
		return;
	case (L_SER):
		strncpy(r_name, "SER\0", 4);
		return;
	case (L_THR):
		strncpy(r_name, "THR\0", 4);
		return;
	case (L_TYR):
		strncpy(r_name, "TYR\0", 4);
		return;
	case (L_TRP):
		strncpy(r_name, "TRP\0", 4);
		return;
	case (L_UNK):
		strncpy(r_name, "UNK\0", 4);
		return;
	case (L_VAL):
		strncpy(r_name, "VAL\0", 4);
		return;
	case (L_HYP):
		strncpy(r_name, "HYP\0", 4);
		return;
	default:
		fprintf(STDERR, "int_to_resname() : ERROR invalid res_int = *%d*\n", res_int);
		exit(1);
	}
}

int resname_to_int(char *r_name)
{
	switch (*r_name) {
		case ('A'):if (!strncmp("ALA", r_name, 3)) {
			return L_ALA;
		} else if (!strncmp("ASP", r_name, 3)) {
			return L_ASP;
		} else if (!strncmp("ARG", r_name, 3)) {
			return L_ARG;
		} else if (!strncmp("ASN", r_name, 3)) {
			return L_ASN;
		} else if (!strncmp("ASX", r_name, 3)) {
			return L_ASX;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('C'):
		if (!strncmp("CYS", r_name, 3)) {
			return L_CYS;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('G'):
		if (!strncmp("GLY", r_name, 3)) {
			return L_GLY;
		} else if (!strncmp("GLU", r_name, 3)) {
			return L_GLU;
		} else if (!strncmp("GLN", r_name, 3)) {
			return L_GLN;
		} else if (!strncmp("GLX", r_name, 3)) {
			return L_GLX;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('H'):
		if (!strncmp("HIS", r_name, 3)) {
			return L_HIS;
		} else if (!strncmp("HYP", r_name, 3)) {
			return L_HYP;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('I'):
		if (!strncmp("ILE", r_name, 3)) {
			return L_ILE;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('L'):
		if (!strncmp("LEU", r_name, 3)) {
			return L_LEU;
		} else if (!strncmp("LYS", r_name, 3)) {
			return L_LYS;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('M'):
		if (!strncmp("MET", r_name, 3)) {
			return L_MET;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('P'):
		if (!strncmp("PHE", r_name, 3)) {
			return L_PHE;
		} else if (!strncmp("PRO", r_name, 3)) {
			return L_PRO;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('S'):
		if (!strncmp("SER", r_name, 3)) {
			return L_SER;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('T'):
		if (!strncmp("THR", r_name, 3)) {
			return L_THR;
		} else if (!strncmp("TYR", r_name, 3)) {
			return L_TYR;
		} else if (!strncmp("TRP", r_name, 3)) {
			return L_TRP;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('U'):
		if (!strncmp("UNK", r_name, 3)) {
			return L_UNK;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	case ('V'):
		if (!strncmp("VAL", r_name, 3)) {
			return L_VAL;
		} else {
			fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
			exit(1);
		}

	default:
		fprintf(STDERR, "resname_to_int() : ERROR invalid residue name *%3s*\n", r_name);
		exit(1);

	}
}

void sort_residue(int *f_order, int **f_val, int c_num)
{
	int ind, act_pos, max_act_pos, tmp;

	max_act_pos = 22;
	while (max_act_pos >= 0) {
		act_pos = -1;
		for (ind = 1; ind <= max_act_pos; ind++)
			if (f_val[f_order[ind - 1]][c_num] < f_val[f_order[ind]][c_num]) {
				act_pos = ind - 1;
				tmp = f_order[act_pos];
				f_order[act_pos] = f_order[ind];
				f_order[ind] = tmp;
			}
		max_act_pos = act_pos;
	}

}

/* Extract the segment whose length is given by "segment_length".
   (Recall that the first and the last residue of a sequence are already excluded.)
 
   sequence_heads[0]: the last entry of seg_freq_tbl[] 
*/
void segment_extraction(int segment_length, struct pe_seg_freq_tbl * seg_freq_tbl,
			int * sequence_heads)
{
	int res_ct;
	int mod_ct, seq_ct;
	int cur_head, nxt_head;
	int i;
	
	for (res_ct = 0; res_ct < sequence_heads[0]; res_ct++) {
		seg_freq_tbl[res_ct].pos = res_ct;
		seg_freq_tbl[res_ct].res_freq = 1;
		seg_freq_tbl[res_ct].code_freq = 1;
	}
	/* Invalidate the last "segment_length"-1 residues */
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		seq_ct = my_protein.model[mod_ct].seq_head;
		cur_head = sequence_heads[seq_ct];
		for (seq_ct++; seq_ct <= my_protein.model[mod_ct].seq_tail;
		     seq_ct++) {
			nxt_head = sequence_heads[seq_ct];
			if (cur_head > 0) {
				for (i = 1; i <= segment_length - 1; i++) {
					res_ct = nxt_head - i;
					if (res_ct < cur_head) break;
					seg_freq_tbl[res_ct].pos = -1;
					seg_freq_tbl[res_ct].res_freq = 0;
					seg_freq_tbl[res_ct].code_freq = 0;
				}
			}
			cur_head = nxt_head;
		}
		nxt_head = sequence_heads[0];
		for (i = 1; i <= segment_length - 1; i++) {
			res_ct = nxt_head - i;
			if (res_ct < cur_head) break;
			seg_freq_tbl[res_ct].pos = -1;
			seg_freq_tbl[res_ct].res_freq = 0;
			seg_freq_tbl[res_ct].code_freq = 0;
		}
		break;	/* Consider only one model per protein */
	}	/* mod_ct */
	return;
}

int sort_by_residue(int ind, int pre_ind, int init_act_pos, int segment_length,
		    struct pe_seg_freq_tbl * seg_freq_tbl, struct pe_res_code * res_codes)
{
	int tbl_ind, pre_tbl_ind;
	int res_ct;
	int act_pos;
	int tmp;
	
	act_pos = init_act_pos;
	/* Sort by residue type */
	pre_tbl_ind = seg_freq_tbl[pre_ind].pos;
	tbl_ind = seg_freq_tbl[ind].pos;
	res_ct = 0;
	while ((res_ct < segment_length - 1)
	       && (res_codes[pre_tbl_ind].type == res_codes[tbl_ind].type)) {
		pre_tbl_ind++;
		tbl_ind++;
		res_ct++;
	}
	if (res_codes[pre_tbl_ind].type > res_codes[tbl_ind].type) {
		act_pos = pre_ind;
		tmp = seg_freq_tbl[act_pos].pos;
		seg_freq_tbl[act_pos].pos
			= seg_freq_tbl[ind].pos;
		seg_freq_tbl[ind].pos = tmp;
	} else if (res_codes[pre_tbl_ind].type == res_codes[tbl_ind].type) {
		/* Sort by residue code */
		pre_tbl_ind = seg_freq_tbl[pre_ind].pos;
		tbl_ind = seg_freq_tbl[ind].pos;
		res_ct = 0;
		while ((res_ct < segment_length - 1) 
		       && (res_codes[pre_tbl_ind].code
			   == res_codes[tbl_ind].code)) {
			pre_tbl_ind++;
			tbl_ind++;
			res_ct++;
		}
		if (res_codes[pre_tbl_ind].code > res_codes[tbl_ind].code) {
			act_pos = pre_ind;
			tmp = seg_freq_tbl[act_pos].pos;
			seg_freq_tbl[act_pos].pos = seg_freq_tbl[ind].pos;
			seg_freq_tbl[ind].pos = tmp;
		}
	}
	return act_pos;
}

/* Sort the codes of the fragment (Bubble sort)      */
/* seg_freq_tbl[].pos:  pointer to res_codes[]      */
/* res_codes[]: (residue name and residue U/D code) */
void sort_fragments(int segment_length, 
		    struct pe_seg_freq_tbl * seg_freq_tbl,
		    struct pe_res_code * res_codes,
		    int * sequence_heads)
{
	int act_pos, max_act_pos;
	int ind, pre_ind;
	
	max_act_pos = sequence_heads[0] - segment_length;
	while (max_act_pos >= 0) {
		act_pos = -1;
		pre_ind = 0;
		while (seg_freq_tbl[pre_ind].pos < 0) pre_ind++;
		ind = pre_ind + 1;
		for (; ind <= max_act_pos; ind++) {
			if (seg_freq_tbl[ind].pos < 0) continue;
			act_pos = sort_by_residue(ind, pre_ind, act_pos, 
						  segment_length, seg_freq_tbl,
						  res_codes);
			pre_ind = ind;
		}
		max_act_pos = act_pos;
	}
	return;
}

/* Sorted by residue type and code */
void calculate_code_frequency(int ind, int segment_length,
			      struct pe_seg_freq_tbl * seg_freq_tbl, struct pe_res_code * res_codes,
			      int * pres_ind, int * pcode_ind)
{
	int res_ind;
	int tbl_ind, res_tbl_ind;
	int code_tbl_ind;
	int res_ct;
	
	res_ind = *pres_ind;
	/* Sorted by residue type */
	res_tbl_ind = seg_freq_tbl[res_ind].pos;
	tbl_ind = seg_freq_tbl[ind].pos;
	res_ct = 0;
	while ((res_ct < segment_length - 1)
	       && (res_codes[tbl_ind].type == res_codes[res_tbl_ind].type)) {
		res_tbl_ind++;
		tbl_ind++;
		res_ct++;
	}
	if ((res_ct == segment_length - 1)
	    && (res_codes[tbl_ind].type == res_codes[res_tbl_ind].type)) {
		/* The same residue */
		seg_freq_tbl[res_ind].res_freq++;
		seg_freq_tbl[ind].res_freq = 0;
		/* Sorted by residue code */
		res_tbl_ind = seg_freq_tbl[res_ind].pos;
		tbl_ind = seg_freq_tbl[ind].pos;
		res_ct = 0;
		while ((res_ct < segment_length - 1)
		       && (res_codes[tbl_ind].code
			   == res_codes[res_tbl_ind].code)) {
			res_tbl_ind++;
			tbl_ind++;
			res_ct++;
		}
		code_tbl_ind = seg_freq_tbl[*pcode_ind + res_ct].pos;
		if ((res_ct == segment_length - 1)
		    && (res_codes[tbl_ind].code
			== res_codes[code_tbl_ind].code)) {
			/* The same code */
			seg_freq_tbl[*pcode_ind].code_freq++;
			seg_freq_tbl[ind].code_freq = 0;
	    } else {
		    /* New code */
		    *pcode_ind = ind;
	    }
	} else {
		/* New residue */
		*pres_ind = ind;
		*pcode_ind = ind;
	}
	return;
}

/* Calculate the frequency of each code */
void calculate_frequency_distribution(int segment_length, 
				      struct pe_seg_freq_tbl * seg_freq_tbl,
				      struct pe_res_code * res_codes,
				      int * sequence_heads)
{
	int max_pos;
	int ind;
	int res_ind;
	int code_ind;

	max_pos = sequence_heads[0] - segment_length;
	res_ind = 0;
	while (seg_freq_tbl[res_ind].pos < 0) res_ind++;
	code_ind = res_ind;
	ind = res_ind + 1;
	for (; ind <= max_pos; ind++) {
		if (seg_freq_tbl[ind].pos < 0) continue;
		calculate_code_frequency(ind, segment_length, seg_freq_tbl,
					 res_codes, &res_ind, &code_ind);
	}
	return;
}

/* Print all codes for the tuple */
void print_tuple_codes(int seg_ct, int res_ind, FILE * fp_stats,
		       int segment_length,
		       struct pe_seg_freq_tbl * seg_freq_tbl,
		       struct pe_res_code * res_codes,
		       struct pe_ud_code ** int_to_code_tab,
		       int * sequence_heads, int print_udcode_flag)
{
	int res_ct;
	int r_code;
	int i;
	int t_per_r;
	
	fprintf(fp_stats, "<%d-tuple id=\"%s", segment_length, 
		res_codes[res_ind].name);
	for (res_ct = 1; res_ct < segment_length; res_ct++)
		fprintf(fp_stats, "-%s", res_codes[res_ind + res_ct].name);
	fprintf(fp_stats, "\" code=\"%8ld", res_codes[res_ind].code);
	for (res_ct = 1; res_ct < segment_length; res_ct++)
		fprintf(fp_stats, "-%8ld", res_codes[res_ind + res_ct].code);
	fprintf(fp_stats, "\" freq=\"%8d", seg_freq_tbl[seg_ct].code_freq);
	if (print_udcode_flag > 0) {
		t_per_r = my_settings.res_code_length;
		fprintf(fp_stats, "\" ud_seq=\"");
		for (res_ct = 0; res_ct < segment_length - 1; res_ct++) {
			r_code = res_codes[res_ind + res_ct].code;
			for (i = 0; i < t_per_r - 1; i++) {
				if (int_to_code_tab[r_code]->ud_seq[i] == C_DOWN)
					fprintf(fp_stats, "D-");
				else
					fprintf(fp_stats, "U-");
			}
			if (int_to_code_tab[r_code]->ud_seq[i] == C_DOWN)
				fprintf(fp_stats, "D ");
			else
				fprintf(fp_stats, "U ");
		}
		r_code = res_codes[res_ind + res_ct].code;
		for (i = 0; i < t_per_r - 1; i++) {
			if (int_to_code_tab[r_code]->ud_seq[i] == C_DOWN)
				fprintf(fp_stats, "D-");
			else
				fprintf(fp_stats, "U-");
		}
		if (int_to_code_tab[r_code]->ud_seq[i] == C_DOWN)
			fprintf(fp_stats, "D");
		else
			fprintf(fp_stats, "U");
	}
	fprintf(fp_stats, "\">\n");
	return;
}

/* sequence_heads[0]: the last entry of seg_freq_tbl[] */
void print_freq_dist(FILE * fp_stats, int segment_length, 
		     struct pe_seg_freq_tbl * seg_freq_tbl,
		     struct pe_res_code * res_codes, struct pe_ud_code ** int_to_code_tab,
		     int * sequence_heads, int print_udcode_flag)
{
	int seg_ct, res_ct;
	int res_ind;
	int t_per_r;
	
	t_per_r = my_settings.res_code_length;
	fprintf(fp_stats, "\n");
	fprintf(fp_stats, "\n");
	fprintf(fp_stats,
"<Frequency distribution of %d-tuples of residues (Endpoints are omitted).>\n",
		segment_length);
	for (seg_ct = 0; seg_ct < sequence_heads[0]; seg_ct++) {
		/* Header for each residue tuple */
		res_ind = seg_freq_tbl[seg_ct].pos;
		if (seg_freq_tbl[seg_ct].res_freq > 0) {
			fprintf(fp_stats, "\n");
			fprintf(fp_stats, "<%d-Tuple id=\"%s", segment_length,
				res_codes[res_ind].name);
			for (res_ct = 1; res_ct < segment_length; res_ct++)
				fprintf(fp_stats, "-%s",
					res_codes[res_ind + res_ct].name);
			fprintf(fp_stats, "\" total_frequency=\"%8d\">\n",
				seg_freq_tbl[seg_ct].res_freq);
		}
		if (seg_freq_tbl[seg_ct].code_freq > 0) {
			print_tuple_codes(seg_ct, res_ind, fp_stats,
					  segment_length, seg_freq_tbl,
					  res_codes, int_to_code_tab,
					  sequence_heads, print_udcode_flag);
		}
	}
	fprintf(fp_stats, "\n");
	fprintf(fp_stats, "</Frequency distribution of residue tuples>\n");
	fflush(fp_stats);
	return;
}

#define EXT2_LENGTH (2+1)	/* ***.ent => ***.stats + '\0' */

/* The frequency distribution of U/D-codes for each segment of residues.

   struct pe_seg_freq_tbl * seg_freq_tbl: seg_freq_tbl[].pos:  pointer to res_codes[]
   struct pe_res_code * res_codes: res_codes[]: (residue name and residue U/D code)
*/
void statsN_print(int segment_length, char * filename,
		   struct pe_seg_freq_tbl * seg_freq_tbl,
		   struct pe_res_code * res_codes,
		   struct pe_ud_code ** int_to_code_tab,
		   int * sequence_heads,
		   int print_udcode_flag)
{
	FILE * fp_stats;
	char * outfile;
	char * s;
	int str_size;

	/* Set outfile to Stats file */
	str_size = (int) strlen(filename);
	outfile = (char *) malloc((str_size + EXT2_LENGTH) * sizeof(char));
	if (outfile == NULL) {
		fprintf(STDERR, 
			"statsN_print() : ERROR outfile[] allocation failed\n");
		exit(1);
	}
	strcpy(outfile, filename);
	s = strrchr(outfile, '.');	/* .ent */
	if ((s == NULL)
	    || ((strcmp(s, ".ent") != 0) && (strcmp(s, ".pdb") != 0))) {	/* 2007-01-16 */
		fprintf(STDERR, "stats3_print(): Input file name error (%s)\n", filename);
		exit(1);
	}
	strcpy(s+1, "stats\0");
	fprintf(STDOUT, "STATS FILE: %s\n", outfile);
	if ((fp_stats = fopen(outfile, "a")) == NULL) {
		fprintf(STDERR, "stats3_print(): can't open %s\n", s);
		exit(1);
	}
	
	segment_extraction(segment_length, seg_freq_tbl, sequence_heads);
	sort_fragments(segment_length, seg_freq_tbl, res_codes, sequence_heads);
	calculate_frequency_distribution(segment_length, seg_freq_tbl,
					 res_codes, sequence_heads);
	print_freq_dist(fp_stats, segment_length, seg_freq_tbl, res_codes,
			int_to_code_tab, sequence_heads, print_udcode_flag);

	if (fclose(fp_stats)) {
		fprintf(STDERR, "statsN_print(): Can't close %s\n", outfile);
		abort();
	}
	if (outfile != NULL) free(outfile);
	return;
}

void print_general_info(char * filename)
{
	FILE *fp_stats;
	FILE *fp_pdb;
	int str_size;
	char *outfile;
	char *s;
	char PDBLine[LINE_WIDTH + 1];

	/* File allocation */
	str_size = (int) strlen(filename);
	outfile = (char *) malloc((str_size + EXT2_LENGTH) * sizeof(char));
	if (outfile == NULL) {
		fprintf(STDERR, 
		"print_general_info() : ERROR outfile[] allocation failed\n");
		exit(1);
	}
	strcpy(outfile, filename);
	s = strrchr(outfile, '.');	/* .ent */
	if ((s == NULL)
	    || ((strcmp(s, ".ent") != 0) && (strcmp(s, ".pdb") != 0))) {	/* 2007-01-16 */
		fprintf(STDERR, 
			"general_info_prin(): Input file name error (%s)\n",
			filename);
		exit(1);
	}
	strcpy(s+1, "stats\0");
	fprintf(STDOUT, "STATS FILE: %s\n", outfile);
	if ((fp_stats = fopen(outfile, "w")) == NULL) {
		fprintf(STDERR, "print_general_info(): Can't open %s\n",
			outfile);
		exit(1);
	}
		    
	/* Load HEADER and COMPND of the PDB file */
	if ((fp_pdb = fopen(filename, "r")) == NULL) {
		fprintf(STDERR, "print_general_info(): Can't open %s\n",
			filename);
		exit(1);
	}
	fprintf(fp_stats,
"----------------------------------------------------------------------------------\n");
	while (fgets(PDBLine, LINE_WIDTH + 1, fp_pdb) != NULL) {
		if (*PDBLine == 'C') {
			if (!strncmp("COMP", PDBLine, 4))
				fprintf(fp_stats, "|%s|\n", PDBLine);
		} else if (*PDBLine == 'H') {
			if (!strncmp("HEAD", PDBLine, 4))
				fprintf(fp_stats, "|%s|\n", PDBLine);
		} else if (*PDBLine == '\n') {
			continue;
		} else {
			break;
		}
	}
	fprintf(fp_stats,
"----------------------------------------------------------------------------------\n");

	/* Print the info */
	fprintf(fp_stats, "  Residue code length = %d",
		my_settings.res_code_length);
	if (my_settings.ca_only_flag == TRUE) {
		if (my_settings.along_backbone == TRUE)
			fprintf(fp_stats, "    (N by N along the backbone)");
		else
			fprintf(fp_stats, "    (N by N)");
	} else {
		fprintf(fp_stats, "    (1 by 1)");
	}
	fprintf(fp_stats, "\n");
	fprintf(fp_stats,
		"  Initial atom = %d-th, Distance between tiles = %8.3f\n",
		my_protein.best_ind, my_protein.ave_dis);
	if (my_settings.ca_only_flag == TRUE)
		    fprintf(fp_stats,
			    "  Average distance between CA atoms = %8.3f\n",
			    my_protein.ave_bond_length);
	else
		    fprintf(fp_stats,
			    "  Average distance between atoms = %8.3f\n",
			    my_protein.ave_bond_length);
	fprintf(fp_stats,
		"  Average error between a tile and a CA atom = %8.3f\n",
		my_protein.ave_error);
	if (0) fprintf(fp_stats, "  Parameters: parameters.model_limit = %d\n",
		my_settings.model_limit);
	fprintf(fp_stats, "\n");
	fprintf(fp_stats, "\n");

	/* PDB data */
	fprintf(fp_stats,
"<PDB_data atoms=\"%d\" residues=\"%d\" sequences=\"%d\" molecules=\"%d\" models=\"%d\">\n",
		my_protein.atm_num, my_protein.res_num, my_protein.seq_num,
		my_protein.mol_num, my_protein.mod_num);

	if (fclose(fp_pdb)) {
		fprintf(STDERR, "print_general_info(): Can't close %s\n",
			filename);
		abort();
	}
	if (fclose(fp_stats)) {
		fprintf(STDERR, "print_general_info(): Can't close %s\n",
			outfile);
		abort();
	}
	return;
}

int compute_each_residue_code(int init_t_ct, int init_tail_ct, int init_res_ct,
			      int seq_ct, struct pe_slt_tile * s_tiles,
			      struct pe_res_code * res_codes,
			      struct pe_ud_code ** int_to_code_tab,
			      int * sequence_heads)
{
	int t_ct, tail_ct, c_ct;
	int res_ct;
	int ind;
	int r_name;
	LONG r_code;
	int mid_point;	/* my_settings.res_code_length <=> 2*mid_point + 1 */
	int t_per_r;
	
	t_per_r = my_settings.res_code_length;
	/* Initial check */
	mid_point = (t_per_r - 1) / 2;
	if ((float) (2 * mid_point + 1) != my_settings.res_code_length) {
		fprintf(STDERR,
			"compute_each_residue_code() : ERROR 2*%d + 1 != %d\n",
			mid_point, my_settings.res_code_length);
		exit(1);
	}

	t_ct = init_t_ct;
	tail_ct = init_tail_ct;
	res_ct = init_res_ct;
	for (; t_ct <= tail_ct; t_ct++) {
		ind = s_tiles[t_ct].error_atom;
		if ((my_protein.atom[ind].atom_name == PE_CA)
		    && (s_tiles[t_ct].atom_flag > 0)) {
			if (sequence_heads[seq_ct] < 0)
				sequence_heads[seq_ct] = res_ct;
			strncpy(res_codes[res_ct].name,
				my_protein.atom[ind].res_name, 3);
			res_codes[res_ct].name[3] = '\0';
			r_name = resname_to_int(my_protein.atom[ind].res_name);
			res_codes[res_ct].type = r_name;
			r_code = 0;
			for (c_ct = t_ct - mid_point; c_ct <= t_ct + mid_point;
			     c_ct++) {
				if (c_ct 
				    < my_protein.sequence[seq_ct].tile_head)
					continue;
				if (c_ct 
				    > my_protein.sequence[seq_ct].tile_tail) {
					r_code = 2 * r_code;
					continue;
				}
				if (s_tiles[c_ct].code == C_DOWN)
					r_code = 2 * r_code;
				else if (s_tiles[c_ct].code == C_UP)
					r_code = 2 * r_code + 1;
				else {
					fprintf(STDERR,
						"compute_each_residue_code() : ERROR s_tiles[%d].code == UNKOWN\n",
						c_ct);
					exit(1);
				}
			}
			if ((int_to_code_tab[r_code]->inv_code < 0)
			    || (t_per_r == 1))
				res_codes[res_ct].code = r_code;
			else
				res_codes[res_ct].code = int_to_code_tab[r_code]
					->inv_code;
			res_ct++;
		}
	}
	return res_ct;
}

/* Consider one model per protein */
void compute_residue_codes(struct pe_slt_tile * s_tiles, 
			   struct pe_seg_freq_tbl * seg_freq_tbl,
			   struct pe_res_code * res_codes,
			   struct pe_ud_code ** int_to_code_tab,
			   int * sequence_heads)
{
	int res_ct;
	int mod_ct, seq_ct;
	int t_ct, tail_ct;
	int ind;

	res_ct = 0;
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			/* Exclude the first and the last residues */
			t_ct = my_protein.sequence[seq_ct].tile_head;
			ind = s_tiles[t_ct].error_atom;
			while (t_ct < my_protein.sequence[seq_ct].tile_tail) {
				if ((my_protein.atom[ind].atom_name == PE_CA)
				    && (s_tiles[t_ct].atom_flag > 0))
					break;
				t_ct++;
				ind = s_tiles[t_ct].error_atom;
			}
			t_ct++;
			tail_ct = my_protein.sequence[seq_ct].tile_tail;
			ind = s_tiles[tail_ct].error_atom;
			while (tail_ct > my_protein.sequence[seq_ct].tile_head){
				if ((my_protein.atom[ind].atom_name == PE_CA)
				    && (s_tiles[tail_ct].atom_flag > 0))
					break;
				tail_ct--;
				ind = s_tiles[tail_ct].error_atom;
			}
			tail_ct--;

			res_ct = compute_each_residue_code(t_ct, tail_ct, 
					res_ct, seq_ct, s_tiles, res_codes,
					int_to_code_tab, sequence_heads);
		}	/* seq_ct */
		break;		/* Consider one model per protein */
	}	/* mod_ct */
	sequence_heads[0] = res_ct;	/* the last entry of seg_freq_tbl[] */

	for (; res_ct < my_protein.res_num; res_ct++) {
		seg_freq_tbl[res_ct].pos = -1;
		seg_freq_tbl[res_ct].res_freq = 0;
		seg_freq_tbl[res_ct].code_freq = 0;
	}

	if (0) {
		fprintf(STDOUT, "\n");
		for (seq_ct = 0; seq_ct <= my_protein.seq_num; seq_ct++) {
			fprintf(STDOUT, "code_statistics_analyzer() : DEBUG sequence_head[%d]= %d\n", seq_ct, sequence_heads[seq_ct]);
		}
	}
	return;
}

/* CA count check */
int check_ca_count(struct pe_slt_tile * s_tiles, int * sequence_heads)
{
	int ca_num, atm_ct, end_ca_num;
	int head_error_atom, tail_error_atom;
	int mod_ct, seq_ct, t_ct;

	ca_num = 0;
	for (atm_ct = 1; atm_ct <= my_protein.atm_num; atm_ct++) {
		if (my_protein.atom[atm_ct].atom_name == PE_CA)
			ca_num++;
	}
	end_ca_num = 0;
	
	for (mod_ct = 1; mod_ct <= my_protein.mod_num; mod_ct++) {
		for (seq_ct = my_protein.model[mod_ct].seq_head;
		     seq_ct <= my_protein.model[mod_ct].seq_tail; seq_ct++) {
			if (my_protein.sequence[seq_ct].atom_head == 0)
				break;	/* no residue */
			t_ct = my_protein.sequence[seq_ct].tile_head;
			while (s_tiles[t_ct].atom_flag == 0) t_ct++;
			head_error_atom = s_tiles[t_ct].error_atom;
			if (my_protein.atom[head_error_atom].atom_name == PE_CA)
				end_ca_num++;
			t_ct = my_protein.sequence[seq_ct].tile_tail;
			while (s_tiles[t_ct].atom_flag == 0) t_ct--;
			tail_error_atom = s_tiles[t_ct].error_atom;
			if ((head_error_atom != tail_error_atom)
			    && (my_protein.atom[tail_error_atom].atom_name
				== PE_CA))
				end_ca_num++;
		}
		break;	/* Consider one model per protein */
	}
	
	if (ca_num != (sequence_heads[0] + end_ca_num) * my_protein.mod_num) {
		fprintf(STDERR,
"check_ca_count() : DEBUG  (ca_num= %d)  !=  ( (seg_freq_tbl size + endpoint_ca_num) * mod_num= (%d + %d) * %d ) \n",
			ca_num, sequence_heads[0], end_ca_num,
			my_protein.mod_num);
		return FAIL;
	}
	if (0) fprintf(STDOUT, "check_ca_count() : DEBUG  ca_num = %d\n\n",
			ca_num);
	return SUCCESS;
}

/* seg_freq_tbl[].pos:  pointer to res_codes[] */
/* res_codes[]: residue's name and U/D code */
/* sequence_heads[]: 0-th entry + size of the sequence */
void code_stats_table_allocation(struct pe_seg_freq_tbl ** pseg_freq_tbl,
				 struct pe_res_code ** pres_codes,
				 int ** psequence_heads)
{
	int res_ct;
	int seq_ct;

	*pseg_freq_tbl = (struct pe_seg_freq_tbl *)
		malloc(my_protein.res_num * sizeof(struct pe_seg_freq_tbl));
	if (*pseg_freq_tbl == NULL) {
		fprintf(STDERR, 
"code_stats_table_allocation() : ERROR seg_freq_tbl[] allocation failed\n");
		exit(1);
	}
	
	*pres_codes = (struct pe_res_code *)
		malloc(my_protein.res_num * sizeof(struct pe_res_code));
	if (*pres_codes == NULL) {
		fprintf(STDERR, 
"code_stats_table_allocation() : ERROR res_code[] allocation failed\n");
		exit(1);
	}
	for (res_ct = 0; res_ct < my_protein.res_num; res_ct++) {
		(*pres_codes)[res_ct].type = -1;
		(*pres_codes)[res_ct].code = 0;
	}

	*psequence_heads = (int *)
		malloc((my_protein.seq_num + 1) * sizeof(int));
	if (*psequence_heads == NULL) {
		fprintf(STDERR, 
"code_stats_table_allocation() : ERROR sequence_heads[] allocation failed\n");
		exit(1);
	}
	for (seq_ct = 0; seq_ct <= my_protein.seq_num; seq_ct++) {
		(*psequence_heads)[seq_ct] = -1;
	}
	return;
}

/*******************************************************************/
/* The frequency distribution of n-tile U/D codes for each residue */
/*******************************************************************/
void code_statistics_analyzer(char * filename, struct pe_slt_tile * s_tiles)
{
	struct pe_seg_freq_tbl *seg_freq_tbl;
	struct pe_res_code *res_codes;
	struct pe_ud_code **int_to_code_tab;
	int *sequence_heads;
	LONG num_of_codes;
	int c_ct;
	
	if (DEBUG) {
		fprintf(STDOUT,
			"code_statistics_analyzer() : start\n");
	}
	code_stats_table_allocation(&seg_freq_tbl, &res_codes, &sequence_heads);

	/* Setup the `r_code --> U/D-sequence conversion' table */
	num_of_codes = (LONG) pow(2.0, (double) my_settings.res_code_length);
	int_to_code_tab = setup_int_to_code_table(num_of_codes);
	if (0) print_int_to_protein_code(int_to_code_tab);
	
	/********************************/
	/* Conpute the code of residues */
	/********************************/
	compute_residue_codes(s_tiles, seg_freq_tbl, res_codes,
			      int_to_code_tab, sequence_heads);
	if (DEBUG) check_ca_count(s_tiles, sequence_heads);

	/*****************************/
	/* Print General Information */
	/*****************************/
	print_general_info(filename);

	/******************************************************************/
	/* The frequency distribution of U/D-codes for each N consecutive */
	/* residues.                                                      */
	/******************************************************************/
	if (my_settings.statsN_ON == TRUE) {
		if (1) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 1);
			statsN_print(1, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 1);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 1);
		}
		if (0) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 2);
			statsN_print(2, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 0);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 2);
		}
		if (1) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 3);
			statsN_print(3, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 0);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 3);
		}
		if (0) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 5);
			statsN_print(5, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 0);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 5);
		}
		if (0) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 7);
			statsN_print(7, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 0);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 7);
		}
		if (0) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 9);
			statsN_print(9, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 0);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 9);
		}
		if (0) {
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  now statsN_print() for %d-tuples running...\n", 11);
			statsN_print(11, filename, seg_freq_tbl, res_codes,
				     int_to_code_tab, sequence_heads, 0);
			fprintf(STDOUT,
"code_statistics_analyzer() : DEBUG  statsN_print() done for %d-tuples.\n", 11);
		}
	}
	/* Memory management */
	for (c_ct = 0; c_ct < num_of_codes; c_ct++)
		if (int_to_code_tab[c_ct]->ud_seq != NULL)
			free(int_to_code_tab[c_ct]->ud_seq);
	if (int_to_code_tab != NULL) free(int_to_code_tab);
	if (seg_freq_tbl != NULL) free(seg_freq_tbl);
	if (res_codes != NULL) free(res_codes);
	if (sequence_heads != NULL) free(sequence_heads);
	return;
}
