/*
 *  pe_step_toward_new_atom.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/10/06.
 *  Copyright 2006, 2007 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern void print_npath_table(void);
extern int seek_previous_target_atom(int ind);
extern int seek_next_target_atom(int ind);
extern float better_dist_from_npath(int t_ind, struct pe_slt_tile * slt_tiles,
				    PE_CODE cur_code, int atm_ct, 
				    PE_DIRECTION direction, int * pbst_path,
				    int * pbst_type);
extern float dist_from_new_tile(int t_ind, struct pe_slt_tile * slt_tiles,
				PE_CODE cur_code, int new_ind,
				PE_DIRECTION direction);
extern void set_cen_of_new_tile_back(int new_ind,
				     struct pe_slt_tile * wrk_tiles,
				     int cur_ind);
extern void dist_from_next_tile_back(int cur_ind,
				     struct pe_slt_tile * wrk_tiles,
				     int atm_ct, float * pnxt_U_error,
				     float * pnxt_D_error);
extern void set_top_of_new_tile_back(int new_ind,
				     struct pe_slt_tile * wrk_tiles,
				     PE_CODE new_code, int cur_ind);
extern void set_cen_of_new_tile_for(int new_ind,
				    struct pe_slt_tile * wrk_tiles, 
				    int cur_ind);
extern void dist_from_next_tile_for(int cur_ind, struct pe_slt_tile * wrk_tiles,
				    int atm_ct, float * pnxt_U_error,
				    float * pnxt_D_error);
extern void set_top_of_new_tile_for(int new_ind, struct pe_slt_tile * wrk_tiles,
				    PE_CODE new_code, int cur_ind);
extern void move_tile_to_target(int t_ind, struct pe_slt_tile * slt_tiles,
				int atm_ct);
extern void rotate_tile_and_trn_table(int t_ind, struct pe_slt_tile * slt_tiles);

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;
extern struct pe_coor_system my_coor;
extern struct pe_trn_table my_trn_table[];
extern struct pe_path_of_length_n ** npath_table;
extern struct pe_path_of_length_n * np_table[NP_TBL_SIZE][24];

/* Check: Are there enough tiles left? */
int available_tile_check(int t_ind, int atm_ct, PE_DIRECTION direction)
{
	if (direction == DIR_FORWARD) {
		if (t_ind + (int) my_settings.tiles_per_res
		    > my_protein.tile_max_num) {
			fprintf(STDERR,
	"available_tile_check() :  ERROR Out of tiles (t_ind=%d, atm_ct=%d, DIR_FORWARD)\n",
				t_ind, atm_ct);
			return FAIL;
		}
	} else {
		if (t_ind - (int) my_settings.tiles_per_res < 1) {
			fprintf(STDERR,
	"available_tile_check() :  ERROR Out of tiles (t_ind=%d, atm_ct=%d, DIR_BACKWARD)\n",
				t_ind, atm_ct);
			return FAIL;
		}
	}
	return SUCCESS;
}

void best_npath_inconsistency_check_forward(int t_ind,
			struct pe_slt_tile * slt_tiles, int atm_ct,
			int bst_path1, int bst_type1, float bst_length1,
			int bst_path2, int bst_type2, float bst_length2,
			int * pbst_path, int * pbst_type, float * pbst_length)
{
	int max_path;
	int bst_path, bst_type;
	float bst_length;
	bst_path = *pbst_path;
	bst_type = *pbst_type;
	bst_length = *pbst_length;
	
	max_path = my_settings.max_path;
	if (((slt_tiles[t_ind].code == C_DOWN) && (max_path / 2 <= bst_path))
	    || ((slt_tiles[t_ind].code == C_UP) && (bst_path < max_path / 2))) {
		if (0) {
			fprintf(STDOUT,
"best_npath_inconsistency_check_forward() : CAUTION slt_tiles[%d].code = %d <=> bst_path = %d, bst_length = %f, max_path = %d\n",
				t_ind, slt_tiles[t_ind].code, bst_path,
				bst_length, max_path);
			fprintf(STDOUT,
"best_npath_inconsistency_check_forward() :       (cur_atm = %d, target atm = %d, cur_tile = %d upward)\n",
				slt_tiles[t_ind].error_atom, atm_ct, t_ind);
			fprintf(STDOUT,
"best_npath_inconsistency_check_forward() :        current %d-th atm  = (%f, %f, %f)\n",
				slt_tiles[t_ind].error_atom,
				my_protein.atom[slt_tiles[t_ind].error_atom].x,
				my_protein.atom[slt_tiles[t_ind].error_atom].y,
				my_protein.atom[slt_tiles[t_ind].error_atom].z);
			fprintf(STDOUT,
"best_npath_inconsistency_check_forward() :        target %d-th atm   = (%f, %f, %f)\n",
				atm_ct, my_protein.atom[atm_ct].x,
				my_protein.atom[atm_ct].y,
				my_protein.atom[atm_ct].z);
		}
		if (slt_tiles[t_ind].code == C_DOWN) {
			bst_length = bst_length1;
			bst_path = bst_path1;
			bst_type = bst_type1;
		} else {
			bst_length = bst_length2;
			bst_path = bst_path2;
			bst_type = bst_type2;
		}
		*pbst_path = bst_path;
		*pbst_type = bst_type;
		*pbst_length = bst_length;
	}
	return;
}

void best_npath_inconsistency_check_backward(int t_ind,
			struct pe_slt_tile * slt_tiles, int atm_ct,
			int bst_path1, int bst_type1, float bst_length1,
			int bst_path2, int bst_type2, float bst_length2,
			int * pbst_path, int * pbst_type, float * pbst_length)
{
	int max_path;
	int bst_path, bst_type;
	float bst_length;
	bst_path = *pbst_path;
	bst_type = *pbst_type;
	bst_length = *pbst_length;
	
	max_path = my_settings.max_path;
	if ((slt_tiles[t_ind].code == C_UP && max_path / 2 <= bst_path)
	    || (slt_tiles[t_ind].code == C_DOWN && bst_path < max_path / 2)) {
		if (0) {
			fprintf(STDOUT,
"best_npath_inconsistency_check_backwar() : CAUTION slt_tiles[%d].code = %d <=> bst_path = %d, bst_length = %f, max_path = %d\n",
				t_ind, slt_tiles[t_ind].code, bst_path,
				bst_length, max_path);
			fprintf(STDOUT,
"best_npath_inconsistency_check_backwar() :       (cur_atm = %d, target atm = %d, cur_tile = %d downward)\n",
				slt_tiles[t_ind].error_atom, atm_ct, t_ind);
			fprintf(STDOUT,
"best_npath_inconsistency_check_backwar() :        current %d-th atm  = (%f, %f, %f)\n",
				slt_tiles[t_ind].error_atom,
				my_protein.atom[slt_tiles[t_ind].error_atom].x,
				my_protein.atom[slt_tiles[t_ind].error_atom].y,
				my_protein.atom[slt_tiles[t_ind].error_atom].z);
			fprintf(STDOUT,
"best_npath_inconsistency_check_backwar() :        target %d-th atm   = (%f, %f, %f)\n",
				atm_ct, my_protein.atom[atm_ct].x,
				my_protein.atom[atm_ct].y,
				my_protein.atom[atm_ct].z);
		}
		if (slt_tiles[t_ind].code == C_UP) {
			bst_length = bst_length1;
			bst_path = bst_path1;
			bst_type = bst_type1;
		} else {
			bst_length = bst_length2;
			bst_path = bst_path2;
			bst_type = bst_type2;
		}
		*pbst_path = bst_path;
		*pbst_type = bst_type;
		*pbst_length = bst_length;
	}
	return;
}

void set_current_tile_forward(int t_ind, struct pe_slt_tile * slt_tiles,
			      int bst_path)
{
	int prv_ind, prv_type;
	int cur_type;
	int i;
	
	prv_ind = t_ind - 1;
	prv_type = slt_tiles[prv_ind].type;
	if (bst_path < my_settings.max_path / 2) {
		slt_tiles[t_ind].code = C_DOWN;
		if (slt_tiles[prv_ind].code == C_DOWN) { /* D -> D */
			slt_tiles[t_ind].type
			= my_trn_table[prv_type].downward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[prv_type].downward_D_top[i]
				    + slt_tiles[prv_ind].top[i];
		} else {	/* U -> D */
			slt_tiles[t_ind].type
			= my_trn_table[prv_type].upward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[prv_type].upward_D_top[i]
				    + slt_tiles[prv_ind].top[i];
		}
		cur_type = slt_tiles[t_ind].type;
		slt_tiles[t_ind].tx = slt_tiles[t_ind].cx
			- my_trn_table[cur_type].cx;
		slt_tiles[t_ind].ty = slt_tiles[t_ind].cy
			- my_trn_table[cur_type].cy;
		slt_tiles[t_ind].tz = slt_tiles[t_ind].cz
			- my_trn_table[cur_type].cz;
	} else {
		slt_tiles[t_ind].code = C_UP;
		if (slt_tiles[prv_ind].code == C_DOWN) { /* D -> U */
			slt_tiles[t_ind].type
				= my_trn_table[prv_type].downward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[prv_type].downward_U_top[i]
				    + slt_tiles[prv_ind].top[i];
		} else {	/* U -> U */
			slt_tiles[t_ind].type
				= my_trn_table[prv_type].upward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[prv_type].upward_U_top[i]
				    + slt_tiles[prv_ind].top[i];
		}
		cur_type = slt_tiles[t_ind].type;
		slt_tiles[t_ind].tx = slt_tiles[t_ind].cx
			- my_trn_table[cur_type].cx;
		slt_tiles[t_ind].ty = slt_tiles[t_ind].cy
			- my_trn_table[cur_type].cy;
		slt_tiles[t_ind].tz = slt_tiles[t_ind].cz
			- my_trn_table[cur_type].cz;
	}
	return;
}

void set_current_tile_backward(int t_ind, struct pe_slt_tile * slt_tiles,
			       int bst_path)
{
	int nxt_ind, nxt_type;
	int cur_type;
	int i;
	
	nxt_ind = t_ind + 1;
	nxt_type = slt_tiles[nxt_ind].type;
	if (bst_path < my_settings.max_path / 2) {
		slt_tiles[t_ind].code = C_UP;	/* = -1*C_DOWN */
		if (slt_tiles[nxt_ind].code == C_DOWN) { /* U <- D */
			slt_tiles[t_ind].type
				= my_trn_table[nxt_type].upward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[nxt_type].upward_D_top[i]
				    + slt_tiles[nxt_ind].top[i];
		} else {	/* U <- U */
			slt_tiles[t_ind].type
				= my_trn_table[nxt_type].downward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[nxt_type].downward_D_top[i]
				    + slt_tiles[nxt_ind].top[i];
		}
		cur_type = slt_tiles[t_ind].type;
		slt_tiles[t_ind].tx = slt_tiles[t_ind].cx
			- my_trn_table[cur_type].cx;
		slt_tiles[t_ind].ty = slt_tiles[t_ind].cy
			- my_trn_table[cur_type].cy;
		slt_tiles[t_ind].tz = slt_tiles[t_ind].cz
			- my_trn_table[cur_type].cz;
	} else {
		slt_tiles[t_ind].code = C_DOWN;	/* = -1*C_UP */
		if (slt_tiles[nxt_ind].code == C_DOWN) { /* D <- D */
			slt_tiles[t_ind].type
				= my_trn_table[nxt_type].upward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[nxt_type].upward_U_top[i]
				    + slt_tiles[nxt_ind].top[i];
		} else {	/* D <- U */
			slt_tiles[t_ind].type
				= my_trn_table[nxt_type].downward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[t_ind].top[i]
				    = my_trn_table[nxt_type].downward_U_top[i]
				    + slt_tiles[nxt_ind].top[i];
		}
		cur_type = slt_tiles[t_ind].type;
		slt_tiles[t_ind].tx = slt_tiles[t_ind].cx
			- my_trn_table[cur_type].cx;
		slt_tiles[t_ind].ty = slt_tiles[t_ind].cy
			- my_trn_table[cur_type].cy;
		slt_tiles[t_ind].tz = slt_tiles[t_ind].cz
			- my_trn_table[cur_type].cz;
	}
	return;
}

void set_next_tile_forward(int t_ind, struct pe_slt_tile * slt_tiles,
			   int cur_type, int bst_path, int t_ct)
{
	float cur_tile_tx, cur_tile_ty, cur_tile_tz;
	int nxt_ind;
	int i;
	
	nxt_ind = t_ind + 1;
	cur_tile_tx = slt_tiles[t_ind].tx;
	cur_tile_ty = slt_tiles[t_ind].ty;
	cur_tile_tz = slt_tiles[t_ind].tz;
	if (slt_tiles[t_ind].code == C_DOWN) {
		/* Set the next ``flat'' tile */
		slt_tiles[nxt_ind].cx = my_trn_table[cur_type].downward_cx
			+ cur_tile_tx;
		slt_tiles[nxt_ind].cy = my_trn_table[cur_type].downward_cy
			+ cur_tile_ty;
		slt_tiles[nxt_ind].cz = my_trn_table[cur_type].downward_cz
			+ cur_tile_tz;
		
		/* Set the next ``slant'' tile */
		if (npath_table[cur_type][bst_path].code[t_ct + 1] == C_DOWN) {
			slt_tiles[nxt_ind].code = C_DOWN;	/* D -> D */
			slt_tiles[nxt_ind].tx
			    = my_trn_table[cur_type].downward_D_tx +cur_tile_tx;
			slt_tiles[nxt_ind].ty
			   = my_trn_table[cur_type].downward_D_ty + cur_tile_ty;
			slt_tiles[nxt_ind].tz
			   = my_trn_table[cur_type].downward_D_tz + cur_tile_tz;
			slt_tiles[nxt_ind].type
			    = my_trn_table[cur_type].downward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i]
				    = my_trn_table[cur_type].downward_D_top[i]
				    + slt_tiles[t_ind].top[i];
		} else {	/* D -> U */
			slt_tiles[nxt_ind].code = C_UP;
			slt_tiles[nxt_ind].tx
			   = my_trn_table[cur_type].downward_U_tx + cur_tile_tx;
			slt_tiles[nxt_ind].ty
			   = my_trn_table[cur_type].downward_U_ty + cur_tile_ty;
			slt_tiles[nxt_ind].tz
			   = my_trn_table[cur_type].downward_U_tz + cur_tile_tz;
			slt_tiles[nxt_ind].type
			    = my_trn_table[cur_type].downward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i]
				    = my_trn_table[cur_type].downward_U_top[i]
				    + slt_tiles[t_ind].top[i];
		}
	} else {	/* if(slt_tiles[t_ind].code == C_DOWN) { */
		/* Set the next ``flat'' tile */
		slt_tiles[nxt_ind].cx = my_trn_table[cur_type].upward_cx
			+ cur_tile_tx;
		slt_tiles[nxt_ind].cy = my_trn_table[cur_type].upward_cy
			+ cur_tile_ty;
		slt_tiles[nxt_ind].cz = my_trn_table[cur_type].upward_cz
			+ cur_tile_tz;
		
		/* Set the next ``slant'' tile */
		if (npath_table[cur_type][bst_path].code[t_ct + 1] == C_DOWN) {
			slt_tiles[nxt_ind].code = C_DOWN;	/* U -> D */
			slt_tiles[nxt_ind].tx
			    = my_trn_table[cur_type].upward_D_tx + cur_tile_tx;
			slt_tiles[nxt_ind].ty
			    = my_trn_table[cur_type].upward_D_ty + cur_tile_ty;
			slt_tiles[nxt_ind].tz
			    = my_trn_table[cur_type].upward_D_tz + cur_tile_tz;
			slt_tiles[nxt_ind].type
			    = my_trn_table[cur_type].upward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i]
					= my_trn_table[cur_type].upward_D_top[i]
					+ slt_tiles[t_ind].top[i];
		} else {	/* U -> U */
			slt_tiles[nxt_ind].code = C_UP;
			slt_tiles[nxt_ind].tx
			    = my_trn_table[cur_type].upward_U_tx + cur_tile_tx;
			slt_tiles[nxt_ind].ty
			    = my_trn_table[cur_type].upward_U_ty + cur_tile_ty;
			slt_tiles[nxt_ind].tz
			    = my_trn_table[cur_type].upward_U_tz + cur_tile_tz;
			slt_tiles[nxt_ind].type
			    = my_trn_table[cur_type].upward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i]
					= my_trn_table[cur_type].upward_U_top[i]
					+ slt_tiles[t_ind].top[i];
		}
	}
	return;
}

void set_next_tile_backward(int t_ind, struct pe_slt_tile * slt_tiles,
			    int cur_type, int bst_path, int t_ct)
{
	float cur_tile_tx, cur_tile_ty, cur_tile_tz;
	int nxt_ind;
	int i;
	
	nxt_ind = t_ind - 1;
	cur_tile_tx = slt_tiles[t_ind].tx;
	cur_tile_ty = slt_tiles[t_ind].ty;
	cur_tile_tz = slt_tiles[t_ind].tz;
	if (slt_tiles[t_ind].code == C_DOWN) {
		/* Set the next ``flat'' tile */
		slt_tiles[nxt_ind].cx = my_trn_table[cur_type].upward_cx
		+ cur_tile_tx;
		slt_tiles[nxt_ind].cy = my_trn_table[cur_type].upward_cy
			+ cur_tile_ty;
		slt_tiles[nxt_ind].cz = my_trn_table[cur_type].upward_cz
			+ cur_tile_tz;
		
		/* Set the previous ``slant'' tile */
		if (npath_table[cur_type][bst_path].code[t_ct + 1] == C_DOWN) {
			slt_tiles[nxt_ind].code = C_UP;	/* U <- D */
			slt_tiles[nxt_ind].tx
				= my_trn_table[cur_type].upward_D_tx
				+ cur_tile_tx;
			slt_tiles[nxt_ind].ty
				= my_trn_table[cur_type].upward_D_ty
				+ cur_tile_ty;
			slt_tiles[nxt_ind].tz
				= my_trn_table[cur_type].upward_D_tz
				+ cur_tile_tz;
			slt_tiles[nxt_ind].type
				= my_trn_table[cur_type].upward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i]
					= my_trn_table[cur_type].upward_D_top[i]
					+ slt_tiles[t_ind].top[i];
		} else {	/* D <- D */
			slt_tiles[nxt_ind].code = C_DOWN;
			slt_tiles[nxt_ind].tx
				= my_trn_table[cur_type].upward_U_tx
				+ cur_tile_tx;
			slt_tiles[nxt_ind].ty
				= my_trn_table[cur_type].upward_U_ty
				+ cur_tile_ty;
			slt_tiles[nxt_ind].tz
				= my_trn_table[cur_type].upward_U_tz
				+ cur_tile_tz;
			slt_tiles[nxt_ind].type
				= my_trn_table[cur_type].upward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i]
					= my_trn_table[cur_type].upward_U_top[i]
					+ slt_tiles[t_ind].top[i];
		}
	} else {	/* if(slt_tiles[t_ind].code == C_DOWN) { */
		/* Set the next ``flat'' tile */
		slt_tiles[nxt_ind].cx = my_trn_table[cur_type].downward_cx
		+ cur_tile_tx;
		slt_tiles[nxt_ind].cy = my_trn_table[cur_type].downward_cy
			+ cur_tile_ty;
		slt_tiles[nxt_ind].cz = my_trn_table[cur_type].downward_cz
			+ cur_tile_tz;
		
		/* Set the next ``slant'' tile */
		if (npath_table[cur_type][bst_path].code[t_ct + 1] == C_DOWN) {
			slt_tiles[nxt_ind].code = C_UP;	/* U <- U */
			slt_tiles[nxt_ind].tx
				= my_trn_table[cur_type].downward_D_tx
				+ cur_tile_tx;
			slt_tiles[nxt_ind].ty
				= my_trn_table[cur_type].downward_D_ty
				+ cur_tile_ty;
			slt_tiles[nxt_ind].tz
				= my_trn_table[cur_type].downward_D_tz
				+ cur_tile_tz;
			slt_tiles[nxt_ind].type
				= my_trn_table[cur_type].downward_D_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i] =
					my_trn_table[cur_type].downward_D_top[i]
					+ slt_tiles[t_ind].top[i];
		} else {	/* D <- U */
			slt_tiles[nxt_ind].code = C_DOWN;	/* -1*C_UP */
			slt_tiles[nxt_ind].tx
				= my_trn_table[cur_type].downward_U_tx
				+ cur_tile_tx;
			slt_tiles[nxt_ind].ty
				= my_trn_table[cur_type].downward_U_ty
				+ cur_tile_ty;
			slt_tiles[nxt_ind].tz
				= my_trn_table[cur_type].downward_U_tz
				+ cur_tile_tz;
			slt_tiles[nxt_ind].type
				= my_trn_table[cur_type].downward_U_type;
			for (i = 0; i < 4; i++)
				slt_tiles[nxt_ind].top[i] =
					my_trn_table[cur_type].downward_U_top[i]
					+ slt_tiles[t_ind].top[i];
		}
	}
	return;
}

int step_ca_result_check(int t_ind_save, struct pe_slt_tile * slt_tiles,
			 int atm_ct, int nxt_ind, int bst_path,
			 float bst_length)
{
	int cur_type;
	float wrk_x, wrk_y, wrk_z, wrk_length;
	
	cur_type = slt_tiles[t_ind_save].type;
	wrk_x = npath_table[cur_type][bst_path].c.x
		- my_trn_table[cur_type].cx + slt_tiles[t_ind_save].cx;
	wrk_y = npath_table[cur_type][bst_path].c.y
		- my_trn_table[cur_type].cy + slt_tiles[t_ind_save].cy;
	wrk_z = npath_table[cur_type][bst_path].c.z
		- my_trn_table[cur_type].cz + slt_tiles[t_ind_save].cz;
	if (FLOAT_NEQ(wrk_x - slt_tiles[nxt_ind].cx, 1.0)
	    || FLOAT_NEQ(wrk_y - slt_tiles[nxt_ind].cy, 1.0)
	    || FLOAT_NEQ(wrk_z - slt_tiles[nxt_ind].cz, 1.0)) {
		fprintf(STDERR,
			"step_ca_result_check() : ERROR slt_tiles[%d]=(%f,%f,%f)\n",
			nxt_ind, slt_tiles[nxt_ind].cx, slt_tiles[nxt_ind].cy,
			slt_tiles[nxt_ind].cz);
		fprintf(STDERR,
			"step_ca_result_check() :       !=  (%f,%f,%f)  (npath_table[%d][%d] )\n",
			wrk_x, wrk_y, wrk_z, cur_type, bst_path);
		fprintf(STDERR,
			"step_ca_result_check() :       target atom[%d] = (%f,%f,%f)\n",
			atm_ct, my_protein.atom[atm_ct].x,
			my_protein.atom[atm_ct].y,
			my_protein.atom[atm_ct].z);
		return FAIL;
	}
	if (my_settings.along_backbone != TRUE) {
		wrk_x = slt_tiles[nxt_ind].cx - my_protein.atom[atm_ct].x;
		wrk_y = slt_tiles[nxt_ind].cy - my_protein.atom[atm_ct].y;
		wrk_z = slt_tiles[nxt_ind].cz - my_protein.atom[atm_ct].z;
		wrk_length = wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z;
		if (FLOAT_NEQ(bst_length - wrk_length, my_coor.ave_dis)) {
			fprintf(STDERR,
				"step_ca_result_check() : ERROR %d-th tile, bst_length= %f  !=  wrk_length= %f\n",
				nxt_ind, bst_length, wrk_length);
			fprintf(STDERR,
				"step_ca_result_check() :       (bst_path = %d  against = %d-th atom",
				bst_path, atm_ct);
			if (my_protein.atom[atm_ct].atom_name == PE_CC)
				fprintf(STDERR, " (C ) )\n");
			else if (my_protein.atom[atm_ct].atom_name == PE_CA)
				fprintf(STDERR, " (CA) )\n");
			else if (my_protein.atom[atm_ct].atom_name == PE_NN)
				fprintf(STDERR, " (N ) )\n");
			else
				fprintf(STDERR, " (UN) )\n");
			
			wrk_x = (npath_table[cur_type][bst_path].c.x
				 - my_trn_table[cur_type].cx
				 + slt_tiles[t_ind_save].cx)
				- my_protein.atom[atm_ct].x;
			wrk_y = (npath_table[cur_type][bst_path].c.y
				 - my_trn_table[cur_type].cy
				 + slt_tiles[t_ind_save].cy)
				- my_protein.atom[atm_ct].y;
			wrk_z = (npath_table[cur_type][bst_path].c.z
				 - my_trn_table[cur_type].cz
				 + slt_tiles[t_ind_save].cz)
				- my_protein.atom[atm_ct].z;
			wrk_length = wrk_x*wrk_x + wrk_y*wrk_y + wrk_z*wrk_z;
			fprintf(STDERR,
				"step_ca_result_check() : ERROR  wrk_length= %f  at %d-th tile, cur_type = %d, bst_path = %d\n",
				wrk_length, t_ind_save, cur_type, bst_path);
			return FAIL;
		}
	}
	if (slt_tiles[nxt_ind].code != 0) {
		fprintf(STDERR,
			"step_ca_result_check() : ERROR (slt_tiles[%d].code = %d) != 0\n",
			nxt_ind, slt_tiles[nxt_ind].code);
		return FAIL;
	}
	return SUCCESS;
}

int step_overflow_check(int t_ind, struct pe_slt_tile * slt_tiles, int prv_tile,
			PE_DIRECTION direction, float cur_error, float next_error)
{
	int flag;
	
	flag = SUCCESS;
	if (direction == DIR_FORWARD) {
		if ((t_ind >= my_protein.tile_max_num)
		    && ((NOT_REACHED_YET(cur_error) && cur_error > next_error)
			|| NOT_OVER_RUN(cur_error - next_error))) {
			flag = FAIL;
		}
	} else {
		if ((t_ind <= 0)
		    && ((NOT_REACHED_YET(cur_error) && cur_error > next_error)
			|| NOT_OVER_RUN(cur_error - next_error))) {
			flag = FAIL;
		}
	}
	if (flag == FAIL) {
		fprintf(STDERR,
			"step_overflow_check() : slt_table[t_ind] overflow (t_ind=%d)\n",
			t_ind);
		if (DEBUG) print_my_slt_tiles(slt_tiles);
		return FAIL;
	} else {
		return SUCCESS;
	}
}

int loop_of_six_tiles_check(int t_ind, struct pe_slt_tile * slt_tiles,
			    int prv_tile, PE_DIRECTION direction)
{
	int j;
	
	if ((prv_tile < my_protein.tile_max_num)
	    && (prv_tile > 0)
	    && (slt_tiles[prv_tile].error == slt_tiles[t_ind].error)
	    && (my_protein.atom[slt_tiles[prv_tile].error_atom].chain_id
		== my_protein.atom[slt_tiles[t_ind].error_atom].chain_id)
	    && (slt_tiles[prv_tile].tx == slt_tiles[t_ind].tx)
	    && (slt_tiles[prv_tile].ty == slt_tiles[t_ind].ty)
	    && (slt_tiles[prv_tile].tz == slt_tiles[t_ind].tz)
	    && (slt_tiles[prv_tile].type == slt_tiles[t_ind].type)){
		fprintf(STDERR,
			"step_toward_new_atom() : ERROR 6-tile loop check (t_ind=%d)\n",
			t_ind);
		fprintf(STDERR,
			"loop_of_six-tiles_ckeckm() :         my_coor.ave_dis * (3.0/settings.tiles_per_res + my_settings.dist_factor)\n");
		fprintf(STDERR,
			"loop_of_six-tiles_ckeckm() :         = %8.3f * (3.0/%8.3f + %8.3f) = %8.3f\n",
			my_coor.ave_dis, my_settings.tiles_per_res, 
			my_settings.dist_factor,
			my_coor.ave_dis * (3.0 / my_settings.tiles_per_res +
					   my_settings.dist_factor));
		if (direction == DIR_FORWARD) {
			for (j = 0; j <= 6; j++) {
				fprintf(STDERR,
					"loop_of_six-tiles_ckeck() :         slt_tiles[%d].error= %8.3f\n",
					prv_tile + j,
					slt_tiles[prv_tile + j].error);
			}
		} else {
			for (j = 0; j <= 6; j++)
				fprintf(STDERR,
					"loop_of_six-tiles_ckeck() :         slt_tiles[%d].error= %8.3f\n",
					prv_tile - j,
					slt_tiles[prv_tile - j].error);
		}
		if (0) print_my_slt_tiles(slt_tiles);
		return FAIL;
	}
	return SUCCESS;
}

int step_ca_initial_check(int t_ind, struct pe_slt_tile * slt_tiles,
			  int atm_ct, PE_DIRECTION direction)
{
	if ((t_ind > my_protein.tile_max_num) || (t_ind <= 0)) {
		fprintf(STDERR,
			"step_ca_initial_check() : t_ind= %d <=> my_protein.tile_max_num= %d (target atm_ct=%d)\n",
			t_ind, my_protein.tile_max_num, atm_ct);
		print_my_slt_tiles(slt_tiles);
		return FAIL;
	}
	if ((atm_ct > my_protein.atm_num) || (atm_ct <= 0)) {
		fprintf(STDERR,
			"step_ca_initial_check() : atm_ct= %d <=> my_protein.atm_num= %d (t_ind=%d)\n",
			atm_ct, my_protein.atm_num, t_ind);
		print_my_slt_tiles(slt_tiles);
		return FAIL;
	}
	if (available_tile_check(t_ind, atm_ct, direction) == FAIL) {
		fprintf(STDERR,
			"step_ca_initial_check() : Out of tiles\n");
		return FAIL;
	}
	return SUCCESS;
}

float compute_best_npath(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct,
			 PE_DIRECTION direction, int * pbst_path,
			 int * pbst_type)
{
	int bst_path, bst_path1, bst_path2;
	int bst_type, bst_type1, bst_type2;
	float bst_length, bst_length1, bst_length2;
	
	bst_length1 = better_dist_from_npath(t_ind, slt_tiles, C_DOWN, atm_ct,
					     direction, &bst_path1, &bst_type1);
	bst_length2 = better_dist_from_npath(t_ind, slt_tiles, C_UP, atm_ct,
					     direction, &bst_path2, &bst_type2);
	if (bst_length1 > bst_length2) {
		bst_length = bst_length2;
		bst_path = bst_path2;
		bst_type = bst_type2;
	} else {
		bst_length = bst_length1;
		bst_path = bst_path1;
		bst_type = bst_type1;
	}
	if (direction == DIR_FORWARD) {
		best_npath_inconsistency_check_forward(t_ind, slt_tiles, atm_ct,
					bst_path1, bst_type1, bst_length1,
					bst_path2, bst_type2, bst_length2,
					&bst_path, &bst_type, &bst_length);
		if (slt_tiles[t_ind].type < 0) {	/* If not assigned */
			set_current_tile_forward(t_ind, slt_tiles, bst_path);
		}
	} else {
		best_npath_inconsistency_check_backward(t_ind, slt_tiles,atm_ct,
					bst_path2, bst_type2, bst_length2,
					bst_path1, bst_type1, bst_length1,
					&bst_path, &bst_type, &bst_length);
		if (slt_tiles[t_ind].type < 0) {	/* If not assigned */
			set_current_tile_backward(t_ind, slt_tiles, bst_path);
		}
	}
	if (bst_type != slt_tiles[t_ind].type) {
		fprintf(STDOUT,
			"compute_best_npath() : ERROR at slt_tiles[%d].type =%d  !=  bst_type = %d\n",
			t_ind, slt_tiles[t_ind].type, bst_type);
		print_my_slt_tiles(slt_tiles);
		exit(1);
	}
	*pbst_path = bst_path;
	*pbst_type = bst_type;
	return bst_length;
}

float compute_best_bond(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct,
			PE_DIRECTION direction, int * pbst_path,
			int * pbst_type)
{
	int i, bst_ind;
	int path_length, bst_path_length;
	int cur_path, bst_path;
	int cur_type, bst_type;
	float cur_error;
	float bst_error;
	
	bst_ind = -1;
	bst_path_length = -1;
	if (my_settings.variable_path_length == TRUE) {
		bst_error = -1.0;
		path_length = (int) my_settings.tiles_per_res - NP_TBL_DISP;
		for (i = 0; i < NP_TBL_SIZE; i++, path_length++) {
			npath_table = np_table[i];
			my_settings.max_path = (int) pow(2.0,
							 (double) path_length);
			cur_error = compute_best_npath(t_ind, slt_tiles, atm_ct,
						       direction, &cur_path,
						       &cur_type);
			if (bst_error < 0 || cur_error < bst_error) {
				bst_ind = i;
				bst_error = cur_error;
				bst_path = cur_path;
				bst_type = cur_type;
				bst_path_length = path_length;
			}
		}
		if (0) fprintf(STDOUT,
			"compute_best_bond() : DEBUG best_ind= %d, cur_error= %f, cur_path= %d, cur_type= %d\n",
			bst_ind, bst_error, bst_path, bst_type);
		npath_table = np_table[bst_ind];
		my_settings.max_path = (int) pow(2.0, (double) bst_path_length);
		my_settings.path_length = bst_path_length;
	} else {
		path_length = (int) my_settings.tiles_per_res;
		my_settings.max_path = (int) pow(2.0, (double) path_length);
		my_settings.path_length = path_length;
		npath_table = np_table[NP_TBL_DISP];
		bst_error = compute_best_npath(t_ind, slt_tiles, atm_ct,
					       direction, &bst_path, &bst_type);
	}
	
	*pbst_path = bst_path;
	*pbst_type = bst_type;
	return bst_error;
}

/* Set n - 2 slant tiles up to the next CA atom
(Current CA=tile0)->(t1)->(t2)-> ... ->(tn-1) --> (Next CA)
where n = path_length
*/
int set_tiles_forward(int init_t_ind, struct pe_slt_tile * slt_tiles,
		      int init_cur_type, int bst_path)
{
	int t_ind, nxt_ind;
	int cur_type;
	int t_ct;
	
	t_ind = init_t_ind;
	cur_type = init_cur_type;
	for (t_ct = 0; t_ct < my_settings.path_length - 1; t_ct++) {
		set_next_tile_forward(t_ind, slt_tiles, cur_type, bst_path,
				      t_ct);
		t_ind++;
		cur_type = slt_tiles[t_ind].type;
	}
	
	/* Set the center of the ``next flat'' CA tile  */
	nxt_ind = t_ind + 1;
	if (slt_tiles[t_ind].code == C_DOWN) {
		slt_tiles[nxt_ind].cx = my_trn_table[cur_type].downward_cx
			+ slt_tiles[t_ind].tx;
		slt_tiles[nxt_ind].cy = my_trn_table[cur_type].downward_cy
			+ slt_tiles[t_ind].ty;
		slt_tiles[nxt_ind].cz = my_trn_table[cur_type].downward_cz
			+ slt_tiles[t_ind].tz;
	} else {
		slt_tiles[nxt_ind].cx = my_trn_table[cur_type].upward_cx
			+ slt_tiles[t_ind].tx;
		slt_tiles[nxt_ind].cy = my_trn_table[cur_type].upward_cy
			+ slt_tiles[t_ind].ty;
		slt_tiles[nxt_ind].cz = my_trn_table[cur_type].upward_cz
			+ slt_tiles[t_ind].tz;
	}
	return t_ind;
}

/* Set n - 2 slant tiles up to the previous CA atom
(Previous CA) <-- ... <-(t2)<-(t1)<-(Current CA=tile0)
where n = path_length
*/
int set_tiles_backward(int init_t_ind, struct pe_slt_tile * slt_tiles,
		       int init_cur_type, int bst_path)
{
	int t_ind, prv_ind;
	int cur_type;
	int t_ct;
	
	t_ind = init_t_ind;
	cur_type = init_cur_type;
	for (t_ct = 0; t_ct < my_settings.path_length - 1; t_ct++) {
		set_next_tile_backward(t_ind, slt_tiles, cur_type, bst_path,
				       t_ct);
		t_ind--;
		cur_type = slt_tiles[t_ind].type;
	}
	
	/* Set the center of the ``previous flat'' CA tile  */
	prv_ind = t_ind - 1;
	if (slt_tiles[t_ind].code == C_UP) {
		slt_tiles[prv_ind].cx = my_trn_table[cur_type].downward_cx
			+ slt_tiles[t_ind].tx;
		slt_tiles[prv_ind].cy = my_trn_table[cur_type].downward_cy
			+ slt_tiles[t_ind].ty;
		slt_tiles[prv_ind].cz = my_trn_table[cur_type].downward_cz
			+ slt_tiles[t_ind].tz;
	} else {
		slt_tiles[prv_ind].cx = my_trn_table[cur_type].upward_cx
			+ slt_tiles[t_ind].tx;
		slt_tiles[prv_ind].cy = my_trn_table[cur_type].upward_cy
			+ slt_tiles[t_ind].ty;
		slt_tiles[prv_ind].cz = my_trn_table[cur_type].upward_cz
			+ slt_tiles[t_ind].tz;
	}
	return t_ind;
}

/* target atom:  my_protein.atom[atm_ct] */
/* current tile:  slt_tiles[t_ind]   */
/* If (direction==C_Backward)                                                      */
/* next tile: slt_tiles[t_ind - 1]    */
/* next target_atom:  my_protein.atom[atm_ct - 1]                             */
/* If (direction==Forward)                                                        */
/* next tile: slt_tiles[t_ind + 1]      */
/* next target_atom:  my_protein.atom[atm_ct + 1]                             */
/* Cf. seek_next_tile() */
/* sqrt_nxt_error: distance between the next tile and the target atom */
/* t_ind: current tile */
/* atm_ct: target atom */
/* seq_ct: current sequence */
float set_next_tile(float sqrt_nxt_error, int t_ind,
		    struct pe_slt_tile * slt_tiles, int atm_ct, int seq_ct,
		    PE_DIRECTION direction)
{
	int nxt_ind;
	int nxt_atm, prv_atm;
	PE_CODE nxt_code;
	float nxt_error;
	float nxt_U_error, nxt_D_error;
	float nxtnxt_U_error, nxtnxt_D_error;
	float new_nxt_error;
	
	nxt_error = sqrt_nxt_error * sqrt_nxt_error;
	if (direction == DIR_BACKWARD) {
		nxt_ind = t_ind - 1;
		slt_tiles[nxt_ind].error = sqrt_nxt_error;
		slt_tiles[nxt_ind].error_atom = atm_ct;
		slt_tiles[nxt_ind].atom_flag = 0;
		
		nxt_atm = seek_next_target_atom(atm_ct);
		if (nxt_atm < 0)
			nxt_atm = atm_ct;
		slt_tiles[nxt_ind].dx = my_protein.atom[nxt_atm].x
			- my_protein.atom[atm_ct].x;
		slt_tiles[nxt_ind].dy = my_protein.atom[nxt_atm].y
			- my_protein.atom[atm_ct].y;
		slt_tiles[nxt_ind].dz = my_protein.atom[nxt_atm].z
			- my_protein.atom[atm_ct].z;
		if (0) {
			fprintf(STDERR,
				"set_next_tile() : slt_tiles[%d].error= %f against = %d-th atom  [backward]\n",
				nxt_ind, slt_tiles[nxt_ind].error,
				slt_tiles[nxt_ind].error_atom);
		}
		set_cen_of_new_tile_back(nxt_ind, slt_tiles, t_ind);
		dist_from_next_tile_back(t_ind, slt_tiles, atm_ct,
					 &nxt_U_error, &nxt_D_error);
		if (ALREADY_REACHED(sqrt_nxt_error)
		    && (atm_ct > my_protein.sequence[seq_ct].atom_head)
		    && (nxt_U_error > nxt_error)
		    && (nxt_D_error > nxt_error)) {	/* Overrun */
		    dist_from_next_tile_back(t_ind, slt_tiles, atm_ct - 1,
					     &nxtnxt_U_error, &nxtnxt_D_error);
			if (nxtnxt_U_error > nxtnxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_UP;/* Backward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_DOWN;/* Backward UP */
			}
		} else {
			if (nxt_U_error > nxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_UP;/* Backward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_DOWN;/* Backward UP */
			}
		}
		set_top_of_new_tile_back(nxt_ind, slt_tiles, nxt_code, t_ind);
	} else {	/* Forward */
		nxt_ind = t_ind + 1;
		slt_tiles[nxt_ind].error = sqrt_nxt_error;
		slt_tiles[nxt_ind].error_atom = atm_ct;
		slt_tiles[nxt_ind].atom_flag = 0;
		prv_atm = seek_previous_target_atom(atm_ct);
		if (prv_atm < 0)
			prv_atm = atm_ct;
		slt_tiles[nxt_ind].dx = my_protein.atom[atm_ct].x
			- my_protein.atom[prv_atm].x;
		slt_tiles[nxt_ind].dy = my_protein.atom[atm_ct].y
			- my_protein.atom[prv_atm].y;
		slt_tiles[nxt_ind].dz = my_protein.atom[atm_ct].z
			- my_protein.atom[prv_atm].z;
		if (0) {
			fprintf(STDERR,
				"set_next_tile() : slt_tiles[%d].error= %f against = %d-th atom [forward]\n",
				nxt_ind, slt_tiles[nxt_ind].error,
				slt_tiles[nxt_ind].error_atom);
		}
		set_cen_of_new_tile_for(nxt_ind, slt_tiles, t_ind);
		dist_from_next_tile_for(t_ind, slt_tiles, atm_ct,
			&nxt_U_error, &nxt_D_error);
		if (ALREADY_REACHED(sqrt_nxt_error)
		    && (atm_ct < my_protein.sequence[seq_ct].atom_tail)
		    && (nxt_U_error > nxt_error)
		    && (nxt_D_error > nxt_error)) {	/* Over run */
			dist_from_next_tile_for(t_ind, slt_tiles, atm_ct + 1,
					&nxtnxt_U_error, &nxtnxt_D_error);
			if (nxtnxt_U_error > nxtnxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_DOWN;/* Forward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_UP;/* Forward UP */
			}
		} else {
			if (nxt_U_error > nxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_DOWN;/* Forward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_UP;/* Forward UP */
			}
		}
		set_top_of_new_tile_for(nxt_ind, slt_tiles, nxt_code, t_ind);
	}	/* if (direction == DIR_BACKWARD) */
	if (0) {
		fprintf(STDERR,
			"set_next_tile() : DEBUG  to slt_tiles[%d] from slt_tiles[%d]\n",
			nxt_ind, t_ind);
	}
	my_protein.tile_num++;
	return sqrt(new_nxt_error);
}

/*
 target atom:  my_protein.atom[atm_ct]
 current ``flat'' tile:  slt_tiles[t_ind]
 Only cx, cy, cz of the current CA tile have been computed.
 ==> tx, ty, tz of the current CA tile and cx, cy, cz of the next CA tile
 are to be computed.
 *
 p_ct = code[0]*2**(n-1) + code[1]*2**(n-2) + ... + code[n-1]
 code[0]==C_DOWN <=> for (p_ct=0; p_ct < max_path/2; p_ct++)
 code[0]==C_UP   <=> for (p_ct=max_path/2; p_ct < max_path; p_ct++)
 */
/* It returns the brand-new CA tile */
/* t_ind: current tile */
/* atm_ct: target atom */
int step_toward_new_ca_atom(int t_ind, struct pe_slt_tile * slt_tiles,
			    int atm_ct, PE_DIRECTION direction)
{
	int t_ind_save;
	int nxt_ind;
	int bst_path, bst_type;
	float bst_length;
	
	if (DEBUG) t_ind_save = t_ind;
	if (DEBUG) {
		if (step_ca_initial_check(t_ind, slt_tiles, atm_ct, direction)
		    == FAIL) 
			exit(1);
	}
	
	bst_length = compute_best_bond(t_ind, slt_tiles, atm_ct, direction,
				       &bst_path, &bst_type);
	if (direction == DIR_FORWARD) {
		if (0) fprintf(STDOUT,
			       "step_toward_new_ca_atom() : DEBUG at slt_tiles[%d] bst_type = %d, bst_path = %d, bst_length = %f (forward)\n",
			       t_ind, bst_type, bst_path, sqrt(bst_length));
		t_ind = set_tiles_forward(t_ind, slt_tiles, bst_type, bst_path);
		nxt_ind = t_ind + 1;
	} else {
		if (0) fprintf(STDOUT,
			       "step_toward_new_ca_atom() : DEBUG at slt_tiles[%d] bst_type = %d, bst_path = %d, bst_length = %f (backward)\n",
			       t_ind, bst_type, bst_path, sqrt(bst_length));
		t_ind = set_tiles_backward(t_ind, slt_tiles, bst_type,bst_path);
		nxt_ind = t_ind - 1;
	}
	
	if (DEBUG) {
		if (step_ca_result_check(t_ind_save, slt_tiles, atm_ct, nxt_ind,
					 bst_path, bst_length) == FAIL) {
			if (direction == DIR_FORWARD)
				fprintf(STDERR,
					"step_toward_new_ca_atom() : DEBUG t_ind = %d, atm_ct = %d (forward)\n",
					t_ind, atm_ct);
			else
				fprintf(STDERR,
					"step_toward_new_ca_atom() : DEBUG t_ind = %d, atm_ct = %d (downward)\n",
					t_ind, atm_ct);
			fflush(STDERR);
			exit(1);
		}
	}
	my_protein.tile_num = my_protein.tile_num
		+ (int) my_settings.path_length;
	slt_tiles[nxt_ind].error_atom = atm_ct;
	slt_tiles[nxt_ind].error = sqrt(bst_length);
	
	if (0) {
		fprintf(STDERR,
			"step_toward_new_ca_atom() : slt_tiles[%d].error= %f against = %d-th atom\n",
			nxt_ind, slt_tiles[nxt_ind].error,
			slt_tiles[nxt_ind].error_atom);
	}
	return nxt_ind;
}

/* target atom:  my_protein.atom[atm_ct]   */
/* seq_ct: current sequence */
/* current tile:  slt_tiles[t_ind]   */
int step_toward_new_atom(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct,
			 int seq_ct, PE_DIRECTION direction)
{
	int flag;
	int prv_tile;	/* for 6-tile loop check */
	float cur_error;
	float next_error; /* distance between the next tile and the target */
		
	if (DEBUG) {
		flag = step_ca_initial_check(t_ind, slt_tiles, atm_ct,
					     direction);
		if (flag == FAIL) exit(1);
	}
	cur_error = dist_from_new_tile(t_ind, slt_tiles, slt_tiles[t_ind].code,
				       atm_ct, direction);
	cur_error = sqrt(cur_error);
	next_error = set_next_tile(cur_error, t_ind , slt_tiles, atm_ct,
				   seq_ct, direction);
	/* Current tile */
	if (direction == DIR_FORWARD) {
		t_ind++;
		prv_tile = t_ind - 6;
	} else {
		t_ind--;
		prv_tile = t_ind + 6;
	}
	flag = step_overflow_check(t_ind, slt_tiles, prv_tile,
				   direction, cur_error, next_error);
	if(flag == FAIL) exit(1);
	
	if (my_settings.move_to_target == TRUE) {
		move_tile_to_target(t_ind, slt_tiles, atm_ct);
		return t_ind;
	}
	if (my_settings.rotate_tile == TRUE) {
		rotate_tile_and_trn_table(t_ind, slt_tiles);
	}
	
	while ((NOT_REACHED_YET(cur_error) 
		&& cur_error > next_error)	/* [TODO] */
		|| NOT_OVER_RUN(cur_error - next_error)) {
		cur_error = next_error;
		next_error = set_next_tile(next_error, t_ind, slt_tiles, atm_ct,
					   seq_ct, direction);
		flag = loop_of_six_tiles_check(t_ind, slt_tiles, prv_tile,
					       direction);
		if(flag == FAIL) exit(1);
		if (direction == DIR_FORWARD) {
			t_ind++;
			prv_tile++;
		} else {
			t_ind--;
			prv_tile--;
		}
		flag = step_overflow_check(t_ind, slt_tiles, prv_tile,
					   direction, cur_error, next_error);
		if(flag == FAIL) {
			fprintf(STDERR,
				"step_toward_new_atom() : ERROR tile overflow at the %d-th atom\n",
				atm_ct);
			exit(1);
		}
		
		if (my_settings.rotate_tile == TRUE) {
			rotate_tile_and_trn_table(t_ind, slt_tiles);
		}
	}
	return t_ind;		/* the brand-new current tile */		
}
